"""This module contains the config model for the inverter."""
from typing import Any, Dict

from pysimmods.model.config import ModelConfig
from pysimmods.model.qgenerator import QControl


class InverterConfig(ModelConfig):
    """Inverter config

    Parameter
    ---------
    params : dict
        Contains configuration parameters of the inverter. See
        *attributes* section for more information.

    Attributes
    ----------
    sn_kva : float
        Nominal apparent power of the inverter in [kVA].
    q_control : str, optional
        Set the mode for the inverter. Can be one of *'p_set'*,
        *'q_set'*, *'cos_phi_set'*, *'pq_set'*, and *'qpset.'* Default
        is *'p_set'*. See :class:`~.Inverter` for more information
        about the different modes.
    cos_phi : float, optional
        Cosinus of the phase angle for *q_control* modes with constant
        *cos_phi*. Default is 0.9.
    inverter_mode : str, optional
        Specifies whether the inverter is *'capacitive'* or
        *'inductive'*. Default is *'capacitive'*.

    """

    def __init__(self, params: Dict[str, Any]):
        super().__init__(params)

        self.s_max_kva = params["sn_kva"]
        self._controls = [
            "p_set",
            "q_set",
            "cos_phi_set",
            "pq_set",
            "qp_set",
        ]
        self.q_control = QControl[
            params.get("q_control", "prioritize_p").upper()
        ]
        self.cos_phi = params.get("cos_phi", 0.95)
        self.inverter_mode = params.get("inverter_mode", "capacitive")



"""This module contains the state model for HVAC."""
from pysimmods.model.state import ModelState


class HVACState(ModelState):
    """State parameters of the HVAC model.

    See :class:`.ModelState` for additional information.

    Parameters
    ----------
    inits : dict
        Contains the initial configuration of this HVAC model. See
        the attributes section for specific information of the HVAC
        model.

    Attributes
    ----------
    mass_kg : float
        Mass of the content in [kg].
    c_j_per_kg_k : float
        Mean specific heat capacity of refer cargo in [Jkg^-K^-1].
    theta_t_deg_celsius : float
        Time dependent temperature of the model in [°C].
    cooling : bool
        If set to True, the model is cooling, i.e., consuming power.
    mode : str
        The operation mode of this model.

    """

    def __init__(self, inits):
        super().__init__(inits)

        self.mass_kg = inits.get("mass_kg", 500)
        self.c_j_per_kg_k = inits.get("c_j_per_kg_k", 2390)
        self.theta_t_deg_celsius = inits.get("theta_t_deg_celsius", 21)
        self.cooling = inits.get("cooling", True)
        try:
            if self.cooling.lower() in (0, "f", "false"):
                self.cooling = False
        except AttributeError:
            # cooling is already a bool
            pass

        self.mode = inits.get("mode", "auto")



from pysimmods.model.model import Model


class Buffer(Model):
    """A buffer subtype model.

    This class provides pre-implemented all required functions so that
    derived models only need to provide the step-function. However,
    those functions can be overwritten if needed.

    This class also provides unified access to set_percent for all
    buffer models. With passive sign convention, the buffer returns
    negative power for generation and positive power for consumption.
    With active sign convention, it is the other way around.

    *set_percent* hides the sign convention. In contrast to generator
    and load, the battery is idle with a set_percent of 50.

    """

    # def get_percent_out(self) -> float:
    #     # TODO:
    #     return 0.5 if self.config.use_decimal_percent else 50

    # def get_percent_in(self):
    #     """Return the percentage value that was set as input.

    #     50 % indicates no operation. Lower values represent discharging
    #     and higher values represent charging in passive sign convention.

    #     Depending on the previous definition, there are some spots that
    #     will be remapped. Values between 50 and 50.01 will be mapped to
    #     50.005 and values between 49.995 and 50 will be mapped to
    #     49.995.

    #     Returns values between 0 and 100 unless the use_decimal_percent
    #     flag is set in the model config.

    #     """
    #     p_kw = self.inputs.p_set_kw
    #     if p_kw is None:
    #         return None

    #     if p_kw * self.config.gsign > 0:
    #         # Generating energy
    #         p_max_kw = self.get_pn_discharge_max_kw()
    #         p_min_kw = self.get_pn_discharge_min_kw()
    #         sign = self.config.gsign

    #     elif p_kw * self.config.lsign > 0:
    #         # Consuming energy
    #         p_max_kw = self.get_pn_charge_max_kw()
    #         p_min_kw = self.get_pn_charge_min_kw()
    #         sign = self.config.lsign

    #     else:
    #         return 0.5 if self.config.use_decimal_percent else 50

    #     decimal = (p_kw - p_min_kw) / (p_max_kw - p_min_kw)
    #     if decimal < 0:
    #         # p_kw too low, turn off
    #         decimal = 0.5
    #     else:
    #         decimal = max(0.0001, decimal)
    #         decimal = 0.5 + (decimal * 0.5 * sign)

    #     return decimal if self.config.use_decimal_percent else decimal * 100

    # def set_percent(self, percentage):
    #     """Set the percentage power.

    #     The definitions from get_percent_in are valid here as well.

    #     """
    #     if self.config.use_decimal_percent:
    #         decimal = max(min(abs(percentage), 1.0), 0.0)
    #     else:
    #         # Internally, decimal percentage is used.
    #         decimal = max(min(abs(percentage), 100.0), 0.0) / 100.0

    #     if decimal == 0.5:
    #         self.inputs.p_set_kw = 0
    #     else:
    #         if decimal < 0.5:
    #             decimal = min(0.49995, decimal)
    #         elif decimal > 0.5:
    #             decimal = max(0.50005, decimal)
    #         psc = self.config.psc

    #         if psc and decimal < 0.5 or not psc and decimal > 0.5:
    #             # Generating energy in psc
    #             p_max_kw = self.get_pn_discharge_max_kw()
    #             p_min_kw = self.get_pn_discharge_min_kw()
    #         else:
    #             # Consuming energy in psc
    #             p_max_kw = self.get_pn_charge_max_kw()
    #             p_min_kw = self.get_pn_charge_min_kw()

    #         decimal = abs((0.5 - decimal) * 2)
    #         self.inputs.p_set_kw = p_min_kw + decimal * (p_max_kw - p_min_kw)

    def get_pn_charge_max_kw(self):
        if hasattr(self.config, "p_charge_max_kw"):
            return self.config.p_charge_max_kw * self.config.lsign
        else:
            return self.config.p_max_kw * self.config.lsign

    def get_pn_charge_min_kw(self):
        if hasattr(self.config, "p_charge_min_kw"):
            return self.config.p_charge_min_kw * self.config.lsign
        else:
            return self.config.p_min_kw * self.config.lsign

    def get_pn_discharge_max_kw(self):
        if hasattr(self.config, "p_discharge_max_kw"):
            return self.config.p_discharge_max_kw * self.config.gsign
        else:
            return self.config.p_max_kw * self.config.gsign

    def get_pn_discharge_min_kw(self):
        if hasattr(self.config, "p_discharge_min_kw"):
            return self.config.p_discharge_min_kw * self.config.gsign
        else:
            return self.config.p_min_kw * self.config.gsign

    def get_pn_min_kw(self):
        """Minimum power of battery in kW.

        The minimum power is the maximum discharging power in passive
        sign convention.

        """
        if self.config.psc:
            return self.get_pn_discharge_max_kw()
        else:
            return self.get_pn_charge_max_kw()

    def get_pn_max_kw(self):
        """Maximum power of battery in kW.

        The maximum power is the maximum charging power in passive sign
        convention.

        """
        if self.config.psc:
            return self.get_pn_charge_max_kw()
        else:
            return self.get_pn_discharge_max_kw()

    def set_p_kw(self, p_kw: float) -> None:
        """Set the target active power value.

        With passive sign convention, positive values indicate charging
        and *p_kw* needs to be between pn_charge_min_kw and
        pn_charge_max_kw. Negative values indicate discharging and
        *p_kw* needs to be between pn_discharge_min_kw and
        pn_discharge_max_kw.

        If *p_kw* is lower than the respective min value, it will be set
        to zero. If it is higher than the respective max value, it will
        be set to the max value.

        """
        if self.config.psc and p_kw < 0 or self.config.asc and p_kw > 0:
            # Discharging
            if p_kw < 0:
                p_max = self.get_pn_discharge_min_kw()
                p_min = self.get_pn_discharge_max_kw()
            else:
                p_min = self.get_pn_discharge_min_kw()
                p_max = self.get_pn_discharge_max_kw()
        elif self.config.psc and p_kw > 0 or self.config.asc and p_kw < 0:
            # Charging
            if p_kw < 0:
                p_max = self.get_pn_charge_min_kw()
                p_min = self.get_pn_charge_max_kw()
            else:
                p_min = self.get_pn_charge_min_kw()
                p_max = self.get_pn_charge_max_kw()
        else:
            self.inputs.p_set_kw = 0
            return

        if abs(p_kw) < abs(p_min) and abs(p_kw) < abs(p_max):
            self.inputs.p_set_kw = 0
        else:
            self.inputs.p_set_kw = max(p_min, min(p_max, p_kw))

    def set_q_kvar(self, q_kvar) -> None:
        self.inputs.q_set_kvar = q_kvar

    def get_p_kw(self) -> float:
        return self.state.p_kw * self.config.lsign



"""This module contains the input model for the inverter."""


from distutils.util import strtobool


class InverterInputs:
    """Inverter inputs.

    This model does not inherit from the :class:`~.ModelInputs`,
    because those inputs are not required for the inverter.

    Attributes
    ----------
    p_in_kw : float
        Incoming (available) active power in [kW]
    p_set_kw : float
        Target active power in [kW]. Only used in the *'p_set'*,
        *'pq_set'*, and *'qp_set'* modes of the inverter.
    q_set_kw : float
        Target reactive power in [kVAr]. Only used in the *'q_set'*,
        *'pq_set'*, and *'qp_set'* modes of the inverter.
    cos_phi_set : float
        Target cos phi between 0 and 1. Only used in the
        *'cos_phi_set'* mode of the inverter.
    inverter_inductive: bool
        If True, the inductive inverter mode will used for the next
        step. Otherwise, capacitive mode is used. If not set, the
        inverter mode from the config is used.
    """

    def __init__(self):
        self.p_in_kw = None
        self.p_set_kw = None
        self.q_set_kvar = None
        self.cos_phi_set = None
        self._inductive = None

    def reset(self):
        self.p_in_kw = None
        self.p_set_kw = None
        self.q_set_kvar = None
        self.cos_phi_set = None
        self._inductive = None

    @property
    def inductive(self):
        if self._inductive:
            return 1
        else:
            return 0

    @inductive.setter
    def inductive(self, val):
        if val is None or isinstance(val, bool):
            self._inductive = val
        elif isinstance(val, str):
            self._inductive = strtobool(val)
        elif val != 0:
            self._inductive = True
        else:
            self._inductive = False



"""This module contains the config model for the Biogas plant."""

import itertools
from typing import List

from pysimmods.model.config import ModelConfig

DEFAULT_SCHEDULE_1: List[float] = (
    [0, 0, 0, 0, 10, 10, 10, 10]
    + [10, 10, 10, 10, 100, 100, 25, 25]
    + [25, 25, 25, 100, 100, 10, 10, 10]
)

DEFAULT_SCHEDULE_2: List[float] = (
    [25, 25, 25, 25, 25, 25, 25, 25]
    + [25, 25, 25, 25, 100, 100, 50, 50]
    + [50, 50, 50, 100, 100, 25, 25, 25]
)


DEFAULT_SCHEDULE_3: List[float] = (
    [0, 0, 0, 0, 10, 10, 10, 10]
    + [10, 10, 10, 10, 100, 100, 50, 50]
    + [50, 50, 50, 100, 100, 10, 10, 10]
)


class BiogasConfig(ModelConfig):
    """Config parameters of the biogas plant model.

    This class captures the configuration parameters for the biogas
    model.

    Parameters
    ----------
    params: dict
        A dictionary containing the configuration parameters. See
        attributes section. The key for each attribute is the same as
        the attribute name, e.g.,::

             {'gas_m3_per_day': 100}

        to set the attribute 'gas_m3_per_day.

    Attributes
    ----------
    gas_m3_per_day: float
        Gas production per day in [m^3].
    cap_gas_m3: float
        Capacity of the gas storage in [m^3].
    gas_fill_min_percent: float
        Lower boundary for the gas storage in [%].
    gas_fill_max_percent: float
        Upper boundary for the gas storage in [%].
    ch4_concentration_percent: float, optional
        Concentration of methane gas in [%]. Defaults to 50.302% and
        should usually not be changed.
    num_chps: int
        Specifies the number of CHP units in this biogas plant.
    pn_stages_kw: list
        All possible combinations of setpoints for the chps, sorted in
        ascending order without duplicates. This attribute is
        calculated automatically and can not be provided.

    """

    def __init__(self, params):
        super().__init__(params)

        self.gas_m3_per_day = params["gas_m3_per_day"]
        self.cap_gas_m3 = params["cap_gas_m3"]
        self.gas_fill_min_percent = params["gas_fill_min_percent"]
        self.gas_fill_max_percent = params["gas_fill_max_percent"]
        self.ch4_concentration_percent = params.get(
            "ch4_concentration", 50.302
        )

        self.num_chps = params["num_chps"]
        stages = [
            params[f"chp{idx}"]["pn_stages_kw"] for idx in range(self.num_chps)
        ]
        self.pn_stages_kw = sorted(
            list(
                set(
                    [abs(sum(tup)) for tup in list(itertools.product(*stages))]
                )
            )
        )
        self.p_max_kw = max(self.pn_stages_kw)
        self.p_min_kw = min(self.pn_stages_kw)

        if self.num_chps == 1:
            default_schedule = DEFAULT_SCHEDULE_1
        elif self.num_chps == 2:
            default_schedule = DEFAULT_SCHEDULE_2
        else:
            default_schedule = DEFAULT_SCHEDULE_3

        self.default_p_schedule: List[float] = [
            self.p_min_kw + dv * self.p_max_kw / 100 for dv in default_schedule
        ]
        self.default_q_schedule: List[float] = [0.0] * 24

        self.default_p_schedule = params.get(
            "default_schedule", self.default_p_schedule
        )



from pysimmods.model.inputs import ModelInputs


class DieselGenInputs(ModelInputs):
    """Captures the input variables of the diesel generator model

    Input variables can be modified from outside the diesel generator
    model before calling the step-method. They are used as an input for
    the calculation of new values for the state variables in the
    step-method.

    """



"""MIDAS upgrade module for the weather data simulator."""
import logging

import pandas as pd
from midas.util.upgrade_module import UpgradeModule
from pysimmods.buffer.batterysim.presets import battery_preset
from pysimmods.consumer.hvacsim.presets import hvac_preset
from pysimmods.generator.biogassim.presets import biogas_preset
from pysimmods.generator.chplpgsystemsim.presets import chp_preset
from pysimmods.generator.dieselsim.presets import diesel_presets
from pysimmods.generator.pvsystemsim.presets import pv_preset

from mosaik.exceptions import ScenarioError

from .analysis.analysis import analyze

LOG = logging.getLogger(__name__)


class PysimmodsModule(UpgradeModule):
    """Pysimmods upgrade module for MIDAS."""

    def __init__(self, name="der"):
        super().__init__(
            module_name="der",
            default_scope_name="midasmv",
            default_sim_config_name="Pysimmods",
            default_import_str=(
                "pysimmods.mosaik.pysim_mosaik:PysimmodsSimulator"
            ),
            default_cmd_str=(
                "%(python)s -m pysimmods.mosaik.pysim_mosaik %(addr)s"
            ),
            log=LOG,
        )

        self.flex_import_str = (
            "pysimmods.mosaik.flex_mosaik:FlexibilitySimulator"
        )
        self.flex_cmd_str = (
            "%(python)s -m pysimmods.mosaik.flex_mosaik %(addr)s"
        )
        self.models = {
            "PV": (
                "Photovoltaic",
                "sgen",
                ["bh_w_per_m2", "dh_w_per_m2", "t_air_deg_celsius"],
                ["p_mw", "q_mvar", "t_module_deg_celsius"],
            ),
            "HVAC": (
                "HVAC",
                "load",
                ["t_air_deg_celsius"],
                ["p_mw", "q_mvar"],
            ),
            "CHP": (
                "CHP",
                "sgen",
                ["day_avg_t_air_deg_celsius"],
                ["p_mw", "q_mvar"],
            ),
            "DIESEL": (
                "DieselGenerator",
                "sgen",
                [],
                ["p_mw", "q_mvar"],
            ),
            "BAT": (
                "Battery",
                "storage",
                [],
                ["p_mw", "q_mvar", "soc_percent"],
            ),
            "Biogas": ("Biogas", "sgen", [], ["p_mw", "q_mvar"]),
        }
        self.sensors = list()
        self.actuators = list()

    def check_module_params(self, module_params):
        """Check the module params and provide default values."""

        module_params.setdefault("start_date", self.scenario.base.start_date)
        module_params.setdefault("cos_phi", self.scenario.base.cos_phi)
        module_params.setdefault("q_control", "prioritize_p")
        module_params.setdefault("inverter_mode", "inductive")
        module_params.setdefault(
            "forecast_horizon_hours",
            self.scenario.base.forecast_horizon_hours,
        )
        module_params.setdefault("pv_is_static_t_module", False)
        module_params.setdefault("provide_flexibilities", False)
        module_params.setdefault(
            "flexibility_horizon_hours",
            self.scenario.base.flexibility_horizon_hours,
        )
        module_params.setdefault(
            "flexibility_horizon_start_hours",
            self.scenario.base.flexibility_horizon_start_hours,
        )
        module_params.setdefault("flexibility_frequency", 1)
        module_params.setdefault("num_schedules", 10)
        module_params.setdefault("unit", "mw")
        module_params.setdefault("use_decimal_percent", False)
        module_params.setdefault("prioritize_setpoint", False)
        module_params.setdefault(
            "provide_forecasts", module_params["provide_flexibilities"]
        )
        module_params.setdefault(
            "enable_schedules", module_params["provide_forecasts"]
        )

    def check_sim_params(self, module_params, **kwargs):
        """Check the params for a certain simulator instance."""
        self.sim_params.setdefault("grid_name", self.scope_name)
        self.sim_params.setdefault("start_date", module_params["start_date"])
        self.sim_params.setdefault("cos_phi", module_params["cos_phi"])
        self.sim_params.setdefault("q_control", module_params["q_control"])
        self.sim_params.setdefault(
            "inverter_mode", module_params["inverter_mode"]
        )
        self.sim_params.setdefault(
            "pv_is_static_t_module", module_params["pv_is_static_t_module"]
        )
        self.sim_params.setdefault(
            "forecast_horizon_hours",
            module_params["forecast_horizon_hours"],
        )
        self.sim_params.setdefault(
            "provide_flexibilities",
            module_params["provide_flexibilities"],
        )
        self.sim_params.setdefault(
            "flexibility_horizon_hours",
            module_params["flexibility_horizon_hours"],
        )
        self.sim_params.setdefault(
            "flexibility_horizon_start_hours",
            module_params["flexibility_horizon_start_hours"],
        )
        self.sim_params.setdefault(
            "flexibility_frequency",
            module_params["flexibility_frequency"],
        )
        self.sim_params.setdefault(
            "num_schedules", module_params["num_schedules"]
        )
        self.sim_params.setdefault("unit", module_params["unit"])
        self.sim_params.setdefault(
            "use_decimal_percent",
            module_params["use_decimal_percent"],
        )
        self.sim_params.setdefault(
            "prioritize_setpoint", module_params["prioritize_setpoint"]
        )
        self.sim_params.setdefault(
            "provide_forecasts",
            self.sim_params["provide_flexibilities"]
            or module_params["provide_forecasts"],
        )
        self.sim_params.setdefault(
            "enable_schedules",
            self.sim_params["provide_forecasts"]
            or module_params["enable_schedules"],
        )
        # self.sim_params.setdefault("mapping", dict())
        # self.sim_params.setdefault("weather_provider_mapping", dict())
        # self.sim_params.setdefault("weather_forecast_mapping", dict())

        self.sim_params.setdefault("seed_max", self.scenario.base.seed_max)
        if self.scenario.base.no_rng:
            self.sim_params["seed"] = self.scenario.create_seed()
        else:
            self.sim_params.setdefault("seed", self.scenario.create_seed())

        if (
            self.sim_params["enable_schedules"]
            or self.sim_params["provide_forecasts"]
            or self.sim_params["provide_flexibilities"]
        ):
            if self.sim_params["cmd"] == "python":
                self.sim_params["import_str"] = self.flex_import_str
            elif self.sim_params["cmd"] == "cmd":
                self.sim_params["import_str"] = self.flex_cmd_str

    def start_models(self):
        """Start all models defined in the mapping of a certain simulator."""
        mapping_key = "peak_mapping"
        if not self.sim_params.get(mapping_key, {}):
            if self.sim_params.get("mapping", {}):
                self.sim_params[mapping_key] = self.sim_params.pop("mapping")
            else:
                self.sim_params[mapping_key] = self.create_default_mapping()

        if not self.sim_params[mapping_key]:
            # No mapping configured
            return

        eid_mapping = self.scenario.create_shared_mapping(
            self, self.sim_params["grid_name"], "eid"
        )
        high = "1.0" if self.sim_params["use_decimal_percent"] else "100.0"
        model_ctrs = {}

        for model, info in self.models.items():
            for bus, entities in self.sim_params[mapping_key].items():
                for name, p_peak_mw in entities:
                    if model != name:
                        continue
                    model_ctrs.setdefault(model, 0)
                    model_key = self.scenario.generate_model_key(
                        self, model.lower(), bus, model_ctrs[model]
                    )
                    params = self.generate_model_params(
                        model_key, model, p_peak_mw
                    )
                    sn_mva = params["params"].get("sn_mva", p_peak_mw)
                    full_id = self.start_model(model_key, info[0], params)
                    p_min, p_max, q_min, q_max = get_space(
                        model.lower(), params, p_peak_mw
                    )

                    eid_mapping[full_id] = {
                        "p_mw": p_peak_mw,
                        "bus": bus,
                        "type": info[1],
                        "sn_mva": sn_mva,
                    }
                    if p_min != p_max:
                        self.actuators.append(
                            create_actuator(full_id, "p_set_mw", p_min, p_max)
                        )
                    if q_min != q_max:
                        self.actuators.append(
                            create_actuator(
                                full_id, "q_set_mvar", q_min, q_max
                            )
                        )
                    self.actuators.append(
                        create_actuator(full_id, "set_percent", 0, high)
                    )
                    for attr in self.models[model][3]:
                        self.sensors.append(create_sensor(full_id, attr))

                    model_ctrs[model] += 1

    def connect(self):
        """Connect the models to existing other models."""

        for model, info in self.models.items():
            self._connect_to_weather(model, info)
            try:
                self._connect_to_grid(model, info)
            except KeyError:
                LOG.warning(
                    "No grid for %s found. Will output only to database "
                    "(if configured).",
                    self.sim_params["sim_name"],
                )

    def _connect_to_weather(self, model, info):
        mapping_key = "peak_mapping"
        wpm_key = "weather_provider_mapping"
        wfm_key = "weather_forecast_mapping"

        if not self.sim_params.get(wpm_key, {}):
            self.sim_params[wpm_key] = self.create_default_wpm()
        self.sim_params.setdefault(wfm_key, {})

        model_ctrs = {}

        for bus, entities in self.sim_params[mapping_key].items():
            for name, _ in entities:
                if model != name:
                    continue

                model_ctrs.setdefault(model, 0)
                model_key = self.scenario.generate_model_key(
                    self, model.lower(), bus, model_ctrs[model]
                )
                weather_key = self.get_weather_model(model, model_ctrs[model])
                try:
                    self.connect_entities(weather_key, model_key, info[2])
                except KeyError:
                    LOG.warning(
                        "Weather mapping missing for %s. The simulation"
                        " will fail if the model requires weather input",
                        model,
                    )

                weather_key = self.get_weather_model(
                    model, model_ctrs[model], True
                )
                if weather_key is not None:
                    fc_attrs = [f"forecast_{a}" for a in info[2]]
                    self.connect_entities(weather_key, model_key, fc_attrs)

                model_ctrs[model] += 1

    def _connect_to_grid(self, model, info):
        mapping_key = "peak_mapping"
        model_ctrs = {}

        for bus, entities in self.sim_params[mapping_key].items():
            for name, _ in entities:
                if model != name:
                    continue

                model_ctrs.setdefault(model, 0)
                model_key = self.scenario.generate_model_key(
                    self, model.lower(), bus, model_ctrs[model]
                )
                grid_entity_key = None
                try:
                    grid_entity_key = self.get_grid_entity(info[1], bus)
                    self.connect_entities(
                        model_key, grid_entity_key, info[3][:2]
                    )
                except ScenarioError as e:
                    LOG.warning(
                        "Encountered scenario error while connecting %s"
                        "to grid entity %s: %s",
                        model_key,
                        grid_entity_key,
                        e,
                    )
                except KeyError as e:
                    LOG.warning(
                        "Entity missing for %s: %s or for grid: %s (%s).",
                        model,
                        model_key,
                        grid_entity_key,
                        e,
                    )
                except ValueError as e:
                    LOG.warning(
                        "Entity missing for %s: %s or for grid: %s (%s).",
                        model,
                        model_key,
                        grid_entity_key,
                        e,
                    )
                model_ctrs[model] += 1

    def connect_to_db(self):
        """Connect the models to db."""
        mapping_key = "peak_mapping"
        db_key = self.scenario.find_first_model("store", "database")[0]

        model_ctrs = dict()
        for model, info in self.models.items():

            for bus, entities in self.sim_params[mapping_key].items():
                for name, _ in entities:
                    if model != name:
                        continue

                    model_ctrs.setdefault(model, 0)
                    model_key = self.scenario.generate_model_key(
                        self, model.lower(), bus, model_ctrs[model]
                    )

                    fc_attrs = info[3]
                    try:
                        self.connect_entities(model_key, db_key, fc_attrs)
                    except ScenarioError:
                        # Only FlexibilitySimulator has target
                        self.connect_entities(model_key, db_key, info[3])

                    model_ctrs[model] += 1

    def generate_model_params(self, model_key, model, p_peak_mw):
        """Load model params and initial configurations."""
        mod_params, mod_inits = get_presets(
            model,
            p_peak_mw,
            q_control=self.sim_params["q_control"],
            cos_phi=self.sim_params["cos_phi"],
            inverter_mode=self.sim_params["inverter_mode"],
            is_static_t_module=self.sim_params["pv_is_static_t_module"],
        )

        self.sim_params.setdefault(model_key, dict())
        self.sim_params[model_key].setdefault("params", mod_params)
        self.sim_params[model_key].setdefault("inits", mod_inits)
        return self.sim_params[model_key]

    def create_default_wpm(self):
        """Create a default weather provider mapping."""
        wprovider = None
        for key, val in self.params["weather_params"].items():
            if isinstance(key, dict):
                try:
                    if len(val["weather_mapping"]["WeatherCurrent"]) > 0:
                        wprovider = key
                        break

                except KeyError:
                    pass

        wpmapping = dict()
        for models in self.sim_params["mapping"].values():
            for (model, _) in models:
                wpmapping.setdefault(model, dict())
                wpmapping[model].setdefault(wprovider, list())
                wpmapping[model][wprovider].append(0)

        return wpmapping

    def get_weather_model(self, model, idx, forecast=False):
        if forecast:
            wmn = "weatherforecast"
            mapping = self.sim_params["weather_forecast_mapping"]
        else:
            wmn = "weathercurrent"
            mapping = self.sim_params["weather_provider_mapping"]
        try:
            model_mapping = mapping[model]
        except KeyError:
            msg = (
                f"No {wmn} mapping for model {model} defined. This may "
                + "result in an error if the model depends on that inputs."
            )
            LOG.debug(msg)

            return None

        if isinstance(model_mapping, dict):
            for model_idx, (name, wpidx) in model_mapping.items():
                if model_idx != idx:
                    continue
                wp_scope = name
                wp_idx = wpidx
                # wp_key = f"_{name}_{wmn}_{wpidx}"

        elif isinstance(model_mapping, list):
            wp_scope = model_mapping[0]
            if isinstance(wp_scope, list):
                try:
                    wp_scope, wp_idx = model_mapping[idx]
                except IndexError:
                    wp_scope, wp_idx = model_mapping[-1]
            else:
                # if isinstance(model_mapping[1], list):
                #     try:
                #         wp_idx = model_mapping[1][idx]
                #     except IndexError:
                #         wp_idx = model_mapping[1][0]
                # else:
                wp_idx = model_mapping[1]
            # wp_key = f"_{name}_{wmn}_{wpidx}"
        else:
            raise ValueError("Weather provider mapping: Unknown format.")

        models = self.scenario.find_models("weather", wp_scope, wmn)
        for key in models:
            if key.endswith(f"_{wp_idx}"):
                return key

        return None

    def get_grid_entity(self, mtype, bus):
        models = self.scenario.find_grid_entities(
            self.sim_params["grid_name"], mtype, endswith=f"_{bus}"
        )
        if models:
            for key in models:
                # Return first match
                return key

        if mtype == "storage":
            # The storage type may not be present in the grid
            # Attach the unit to load instead
            return self.get_grid_entity("load", bus)

        # self.logger.info(
        #     "Grid entity for %s, %s at bus %d not found",
        #     self.sim_params["grid_name"],
        #     mtype,
        #     bus,
        # )
        raise ValueError(
            f"Grid entity for {self.sim_params['grid_name']}, {mtype} "
            f"at bus {bus} not found!"
        )

    def get_sensors(self):
        for sensor in self.sensors:
            self.scenario.sensors.append(sensor)
            LOG.debug("Created sensor entry %s.", sensor)

    def get_actuators(self):
        for actuator in self.actuators:
            self.scenario.actuators.append(actuator)
            LOG.debug("Created actuator entry %s.", actuator)

    def download(
        self,
        data_path: str,
        tmp_path: str,
        load_if_necessary: bool,
        force: bool,
    ):
        # No downloads
        pass

    def analyze(
        self,
        name: str,
        data: pd.HDFStore,
        output_folder: str,
        start: int,
        end: int,
        step_size: int,
        full: bool,
    ):
        analyze(name, data, output_folder, start, end, step_size, full)


def create_default_mapping():
    return {
        3: [("PV", 6)],
        4: [("PV", 2)],
        8: [("PV", 2)],
        11: [("PV", 3)],
    }


def get_presets(model, p_peak_mw, **kwargs):
    """Return presets for *model* with *p_peak_mw*.

    The presets are taken from the pysimmods package itself.

    """

    if model == "PV":
        params, inits = pv_preset(p_peak_kw=p_peak_mw * 1e3, **kwargs)
        params["sn_mva"] = params["inverter"]["sn_kva"] * 1e-3
        return params, inits
    elif model == "HVAC":
        return hvac_preset(pn_max_kw=p_peak_mw * 1e3, **kwargs)
    elif model == "CHP":
        return chp_preset(p_kw=p_peak_mw * 1e3, **kwargs)
    elif model == "BAT":
        return battery_preset(pn_max_kw=p_peak_mw * 1e3)
    elif model == "DIESEL":
        return diesel_presets(p_max_kw=p_peak_mw * 1e3)
    elif model == "Biogas":
        return biogas_preset(pn_max_kw=p_peak_mw * 1e3)
    else:
        raise ValueError(f"Model {model} is unknown.")


def create_actuator(
    full_id: str, attr: str, vmin: float, vmax: float, dtype: str = "np.double"
):
    return {
        "actuator_id": f"{full_id}.{attr}",
        "action_space": (
            f"Box(low={vmin}, high={vmax}, shape=(1,), dtype={dtype})"
        ),
    }


def create_sensor(
    full_id: str,
    attr: str,
    vmin: float = -10000,
    vmax: float = 10000,
    dtype: str = "np.double",
):
    return {
        "sensor_id": f"{full_id}.{attr}",
        "observation_space": (
            f"Box(low={vmin}, high={vmax}, shape=(1,), dtype={dtype})"
        ),
    }


def get_space(model, params, p_peak_mw):
    p_min = q_min = q_max = 0
    p_max = p_peak_mw

    if model.lower() in ["pv"]:
        q_min = -params["params"]["sn_mva"]
        q_max = params["params"]["sn_mva"]

    if model.lower() in ["bat"]:
        p_min = -params["params"]["p_discharge_max_kw"] * 1e-3
        p_max = params["params"]["p_charge_max_kw"] * 1e-3

    return p_min, p_max, q_min, q_max



"""This module contains the input model for the chp system."""
from pysimmods.model.inputs import ModelInputs

# from pysimmods.generator.chplpgsim.inputs import CHPLPGInputs
# from pysimmods.other.heatdemandsim.inputs import HeatDemandInputs


class CHPLPGSystemInputs(ModelInputs):
    """captures the inputs of the chp system"""

    def __init__(self):
        super().__init__()

        self.day_avg_t_air_deg_celsius = None
        """Average temperature of the current day in [°C]."""

        self.e_th_demand_set_kwh = None



"""This module contains the state model of the battery."""
from typing import Any, Dict

from ...model.state import ModelState


class BatteryState(ModelState):
    """Captures the state variables of the battery model

    State variables normally change during the simulation process. They are
    updated in the step-method of the battery model.

    There is two types of state variables. State variables like *soc_percent*
    which are used in the recalculation of the state variables in the
    step-method of the battery model and such state variables like *p_kw*
    whose current value doesn't affect the update of state variables in the
    step-method.

    The former require the assignment of intial values. Therefore a dictionary
    with intial values for this kind of state variables has to be passed to the
    constructor of the State class. State variables which are not needed as an
    input for calculating the next state in the step-method are initialized
    with None by default.

    Attributes
    ----------
    soc_percent : float
        Current state of charge of battery in percent of capacity.
    eta_percent : float
        Current efficiency of battery in percent.
    _energy_kwh : float
        Current energy of the battery, calculated during the step in
        [kWh].

    """

    def __init__(self, inits: Dict[str, Any]):
        super().__init__(inits)

        self.soc_percent: float = inits["soc_percent"]
        self.eta_percent: float = None
        self._energy_kwh: float = None



from typing import List
from pysimmods.model.config import ModelConfig

DEFAULT_SCHEDULE: List[float] = (
    [50, 50, 50, 50, 50, 100, 100, 100]
    + [100, 50, 50, 50, 100, 100, 50, 50]
    + [50, 100, 100, 100, 100, 50, 50, 50]
)


class DieselGenConfig(ModelConfig):
    def __init__(self, params):
        super().__init__(params)

        self.p_max_kw = params["p_max_kw"]
        self.p_min_kw = 0.1 * params["p_max_kw"]
        self.q_max_kvar = params["q_max_kvar"]
        self.q_min_kvar = params["q_min_kvar"]

        self.default_p_schedule: List[float] = [
            self.p_min_kw + dv * self.p_max_kw / 100 for dv in DEFAULT_SCHEDULE
        ]
        self.default_q_schedule: List[float] = [0.0] * 24

    # @property
    # def p_max_kw(self):
    #     if self.psc:
    #         return self._p_min_kw * self.gsign
    #     else:
    #         return self._p_max_kw * self.gsign

    # @property
    # def p_min_kw(self):
    #     if self.psc:
    #         return self._p_max_kw * self.gsign
    #     else:
    #         return self._p_min_kw * self.gsign

    # @property
    # def q_max_kvar(self):
    #     return self._q_max_kvar

    # @property
    # def q_min_kvar(self):
    #     return self._q_min_kvar



"""This module contains the input model for the Biogas plant."""
from pysimmods.model.inputs import ModelInputs


class BiogasInputs(ModelInputs):
    """Input parameters for the biogas plant model.

    The biogas model currently requires only general inputs, see
    :class:`.ModelInputs`.

    """



"""This module contains the :class:`.Schedule` that is used by the
:class:`.ForecastModel` and :class:`.FlexibilityModel`.

"""
import logging
from datetime import datetime, timedelta, timezone
from typing import Optional

import numpy as np
import pandas as pd
from pysimmods.util.date_util import GER

LOG = logging.getLogger(__name__)


class Schedule:
    """Schedule class for the :class:`.ScheduleModel`.

    Calling the schedule obj like this::

        df = schedule()

    returns the data frame that stores the schedule.


    Parameters
    ----------

    """

    def __init__(
        self,
        step_size: int = 1,
        horizon_hours: float = 1,
        start_date: Optional[datetime] = None,
        p_name: str = "p_kw",
        q_name: str = "q_kvar",
        p_set_name: Optional[str] = None,
        q_set_name: Optional[str] = None,
    ):
        self._data: pd.DataFrame

        self.step_size = step_size
        self.horizon_hours = horizon_hours
        self.now_dt = start_date

        self.p_name = p_name
        self.q_name = q_name
        self.p_set_name = p_set_name
        self.q_set_name = q_set_name

    def init(self):
        """Initialize the schedule data frame.

        After initialization, the schedule's data frame should have
        three columns *target*, *p_kw*, and *q_kvar* and a number of
        rows, each value initialized with np.nan.

        The number of rows is defined by the
        :attr:`_forecast_horizon_hours` as seconds divided by the
        :attr:`step_size`.

        """
        LOG.debug(
            "Creating new schedule dataframe (and wiping any existing data)."
        )
        columns = [self.p_name, self.q_name]

        if self.p_set_name is None:
            unit = self.p_name.rsplit("_", 1)[1]
            self.p_set_name = f"p_set_{unit}"
        if self.q_set_name is None:
            unit = self.q_name.rsplit("_", 1)[1]
            self.q_set_name = f"q_set_{unit}"

        columns.insert(0, self.q_set_name)
        columns.insert(0, self.p_set_name)

        if self.now_dt is not None:
            index = pd.date_range(
                self.now_dt,
                self.now_dt
                + timedelta(hours=self.horizon_hours)
                - timedelta(seconds=self.step_size),
                freq=f"{self.step_size}S",
            )
        else:
            index = None

        self._data = pd.DataFrame(columns=columns, index=index)
        self._data.index = pd.to_datetime(self._data.index)

    def update(self, other):
        """Update this schedules' data with another dataframe.

        Parameters
        ----------
        other : pandas.DataFrame
            A dataframe with datetime as index and values for the
            columns *"target"*, *"p_kw"*, and *"q_kvar"*. Note that
            :attr:`pname` and :attr:`qname` match in both data frames.

        """

        if self._data is None:
            self.init()

        if isinstance(other, Schedule):
            other = other._data

        for col in other.columns:
            if col not in self._data.columns:
                raise ValueError(
                    f"Column '{col}' from other schedule is too much."
                )

        for index, _ in other.iterrows():
            if not self.has_index(index):
                break

        self._data.update(other.loc[:index])
        self._data = pd.concat([self._data, other.loc[index:]])
        self._data = self._data[~self._data.index.duplicated()]
        self._data.index = pd.to_datetime(self._data.index, utc=True)

    def update_row(
        self,
        index: datetime,
        p_set: float,
        q_set: Optional[float],
        p_val: float,
        q_val: float,
    ):

        if index in self._data.index:
            self._data.loc[index][self.p_set_name] = p_set
            self._data.loc[index][self.q_set_name] = q_set
            self._data.loc[index][self.p_name] = p_val
            self._data.loc[index][self.q_name] = q_val

        else:
            tmp_df = pd.DataFrame(
                data={
                    self.p_set_name: p_set,
                    self.q_set_name: q_set,
                    self.p_name: p_val,
                    self.q_name: q_val,
                },
                index=[index],
            )

            self._data = pd.concat([self._data, tmp_df])

        self._data.sort_index(inplace=True)
        self._data.index = pd.to_datetime(self._data.index)
        self._data = self._data.fillna(value=np.nan)

    def update_entry(self, index, col, val):
        if col not in self._data.columns:
            raise ValueError(
                f"Invalid column '{col}'. Supported columns are "
                f"{self._data.columns}."
            )
        if index not in self._data.index:
            self._data = pd.concat(
                [self._data, pd.DataFrame({col: val}, index=[index])]
            )
        self._data.loc[index][col] = val
        self._data.sort_index(inplace=True)
        self._data = self._data.fillna(value=np.nan)

    def has_index(self, index):
        return index in self._data.index

    def reschedule_required(self):
        """Check if a reschedule is required.

        The current schedule is checked for the next hours, specified
        by :attr:`_forecast_horizon_hours*. A reschedule is required
        when on of the indices is missing of one of the values within
        the limit is np.nan.

        Returns
        -------
        bool
            *True* when a reschedule is required and *False* otherwise.
        """
        now = self.now_dt + timedelta(seconds=self.step_size)
        limit = int(self.horizon_hours * 3_600 / self.step_size)

        for _ in range(limit):
            if now not in self._data.index:
                return True
            if self.get(now, self.p_set_name) is None:
                return True
            if self.get(now, self.q_set_name) is None:
                return True
            if self.get(now, self.p_name) is None:
                return True
            elif self.get(now, self.q_name) is None:
                return True

            now += timedelta(seconds=self.step_size)

        return False

    def prune(self):
        self._data = self._data.loc[self.now_dt :]

    def get(self, now, col) -> Optional[float]:
        """Return the value for a certain column.

        If no time index is provided, the current datetime object of
        the schedule is used.

        Returns
        -------
        np.float
            The value of the specified column at index *now*.

        """
        if now is None:
            now = self.now_dt
        elif isinstance(now, str):
            now = datetime.strptime(now, GER).astimezone(timezone.utc)

        if now not in self._data.index:
            return None
        val = self._data.loc[now][col]

        if np.isnan(val):
            return None
        else:
            return val

    def to_json(
        self,
        start_dt: Optional[datetime] = None,
        end_dt: Optional[datetime] = None,
    ) -> str:
        if start_dt is None:
            start_dt = self._data.index[0]

        if end_dt is None:
            end_dt = self._data.index[-1]

        return self._data.loc[start_dt:end_dt].to_json()

    def from_json(self, schedule_json):
        schedule_df = pd.read_json(schedule_json).tz_localize("UTC")
        self.from_dataframe(schedule_df)

    def __call__(self):
        return self._data

    def from_dataframe(self, dataframe: pd.DataFrame):
        self._data = dataframe
        self._data.index = pd.to_datetime(self._data.index)
        if "target" in self._data.columns:
            self.use_absolute_setpoints = False
            self.p_set_name = None
            self.q_set_name = None
        else:
            self.use_absolute_setpoints = True

        for col in self._data.columns:
            if "p_set" in col:
                self.p_set_name = col
            elif "q_set" in col:
                self.q_set_name = col
            elif "p_" in col:
                self.p_name = col
            elif "q_" in col:
                self.q_name = col

        self.now_dt = self._data.index[0].to_pydatetime()
        self.step_size = max(
            1,
            int(
                (self._data.index[-1] - self._data.index[0]).total_seconds()
                / (len(self._data.index) - 1)
            ),
        )

    def __repr__(self):
        return self._data.__repr__()

    def is_empty(self):
        if not hasattr(self, "_data"):
            return True

        return self._data.empty



"""This module contains the heat storage model that is used by the
CHP LPG.
"""
import math


C_WATER = 4.18
"""Specific heat capacity of water in [kJ/(kg*K)]"""


class HeatStorage:
    """Model of a heat storage.

    Parameters
    ----------
    params : dict
        A *dict* containing values for all attributes. The keys all
        have the prefix "storage_".
    inits : dict
        A *dict* containing state variables for the storage. The keys
        all have the prefix "storage_".

    Attributes
    ----------
    cap_l : float
        Capacity of the buffer storage in [l].
    consumption_kwh_per_day : float
        Average loss of energy in [kWh/day].
    t_min_c : float
        Minimal water temperature of the buffer storage in [°C].
    t_max_c : float
        Maximal water temperature of the buffer storage in [°C].
    t_c : float
        Current water temperature of the buffer storage in [°C].
    env_t_c : float, optional
        Current temperature of the environment in [°C]. The storage
        is assumed to be placed indoor and, therefore, defaults to
        19.0 °C.

    """

    def __init__(self, params, inits):

        # Config
        self.cap_l = params["storage_cap_l"]
        self.consumption_kwh_per_day = params[
            "storage_consumption_kwh_per_day"
        ]
        self.t_min_c = params["storage_t_min_c"]
        self.t_max_c = params["storage_t_max_c"]
        self.env_t_c = params.get("storage_env_t_c", 19.0)

        # State
        self.t_c = inits["storage_t_c"]
        self._t_chilled = None

    def _calculate_chilled_t(self):
        """Calculate the chilled temperature.

        This function requires that the :attr:`.env_t_c` is set.

        Returns
        -------
        float
            The chilled temperature

        """

        chill = (
            self.consumption_kwh_per_day
            * 3_600_00
            / (C_WATER * 1_000 * self.cap_l * 24)
        )
        t_next = self.t_c - chill

        # Prevent math domain/division by zero errors
        dif_t_next_env = max(t_next - self.env_t_c, 1e-3)
        dif_t_init_env = max(self.t_c - self.env_t_c, 1e-3)
        div_t = max(dif_t_next_env / dif_t_init_env, 1e-3)

        chill_coef = math.log(div_t) / t_next
        t_chilled = (self.t_c - self.env_t_c) * math.pow(
            math.e, chill_coef
        ) + self.env_t_c

        return t_chilled

    def get_absorbable_energy(self) -> float:
        """Return the amount of energy the storage can absorb."""
        e_th_in_max_kwh = (
            C_WATER * self.cap_l * (self.t_max_c - self.t_c) / 3_600
        )

        return e_th_in_max_kwh

    def get_available_energy(self) -> float:
        """Return the amount of energy the storage has absorbed."""

        e_th_in_min_kwh = (
            C_WATER * self.cap_l * max(0, self.t_c - self.t_min_c) / 3_600
        )

        return e_th_in_min_kwh

    def absorb_energy(self, e_th_prod, e_th_demand):

        # 1 kWh = 3600 kJ
        e_th_delta_kj = (e_th_prod + e_th_demand) * 3_600

        t_chilled = self._calculate_chilled_t()
        t_new = e_th_delta_kj / (C_WATER * self.cap_l) + t_chilled

        self.t_c = max(t_new, self.env_t_c)



"""
This module contains a household class that simulates thermal power demand
for the chp.

"""
import math
import random
from copy import deepcopy
from datetime import timedelta

from pysimmods.other.heatdemandsim import (
    HeatDemandConfig,
    HeatDemandInputs,
    HeatDemandState,
    hprofiles,
)

DEMAND_PROFILES = dict()


class HeatDemand:
    """A simple model for thermal power demand.

    :constant CHP_AVERAGE_RUNTIME:
        source: 'IHK Projekte Hannover, Blockheizkraftwerke,
        Seite 5 IHK Projekte Durschnittliche Betriebsstunden
        eines BHKWs'
    :constant CHP_MIN_TH_CONS_COEF:
        source: 'IHK Projekte Hannover, Blockheizkraftwerke,
        Seite 5 Prozentzahl für die untere minimale thermische
        Leistung des Jahresverbrauchs eines Householdtypes'
    :constant CHP_MAX_TH_CONS_COEF:
        source: 'IHK Projekte Hannover, Blockheizkraftwerke,
        Seite 5 Prozentzahl für die obere maximale thermische
        Leistung des Jahresverbrauchs eines Householdtypes'
    :constant DEGREE_OF_EFFIENCY:
        source: 'Niedertemperatur- und Brennwertkessel.
        Wissenswertes über moderne Zentralheizungsanlagen.
        Hessisches Ministerium für Umwelt, Energie,
        Landwirtschaft und Verbraucherschutz. 2011'

    """

    CHP_AVERAGE_RUNTIME = 5_000.0
    CHP_MIN_TH_CONS_COEF = 1.0
    CHP_MAX_TH_CONS_COEF = 1.5
    # DEGREE_OF_EFFIENCY = 0.93

    def __init__(self, config, state):
        self.config = HeatDemandConfig(config)
        self.state = HeatDemandState(state)
        self.inputs = HeatDemandInputs()

    def step(self):
        """perform a simulation step"""

        next_state = deepcopy(self.state)

        # calculate allocation temperature for the current day
        # (weighted mean temperature of the current and the last
        # three days)
        t_alloc = (
            1.0 * self.inputs.day_avg_t_air_deg_celsius
            + 0.5 * self.state.t_last_1_deg_celsius
            + 0.25 * self.state.t_last_2_deg_celsius
            + 0.125 * self.state.t_last_3_deg_celsius
        ) / (1 + 0.5 + 0.25 + 0.125)

        # profile function 'SigLinDe': calculate daily consumption
        # depending on the allocation temperature. Has both a sigmoid
        # and a linear part
        tmp = math.pow(
            self.config.const_b / (t_alloc - self.config.const_v_0),
            self.config.const_c,
        )
        sigmoid = self.config.const_a / (1.0 + tmp) + self.config.const_d
        linear = max(
            self.config.const_m_h * t_alloc + self.config.const_b_h,
            self.config.const_m_w * t_alloc + self.config.const_b_w,
        )

        daily_cons = (
            self.config.consumer_const
            * (sigmoid + linear)
            * self.config.weekday_const[self.state.now_dt.weekday()]
        )

        next_state.e_th_kwh = (
            daily_cons
            * self.config.degree_of_efficiency
            * self.config.load_profile[self.state.now_dt.hour]
            * self.inputs.step_size
            / 3_600
        )

        next_state.now_dt += timedelta(seconds=self.inputs.step_size)
        if next_state.now_dt.day > self.state.now_dt.day:
            # we reached a new day
            next_state.t_last_3_deg_celsius = next_state.t_last_2_deg_celsius
            next_state.t_last_2_deg_celsius = next_state.t_last_1_deg_celsius
            next_state.t_last_1_deg_celsius = (
                self.inputs.day_avg_t_air_deg_celsius
            )

        self.state = next_state

    @property
    def e_th_kwh(self):
        """Returns the current thermal power demand"""
        return self.state.e_th_kwh

    def get_state(self) -> dict:
        """Return the current state of the model.

        Returns
        -------
        dict
            The current state of the model in form of a dictionary
            containing entries for all state variables. Returned dict
            can be assigned to the *inits* argument when creating a new
            model instance.

        """

        try:
            return {
                attr: getattr(self.state, attr)
                for attr in self.state.__slots__
            }
        except AttributeError:
            return deepcopy(self.state.__dict__)

    def set_state(self, state: dict) -> None:
        """Set the current state of the model.

        Parameters
        ----------
        state : dict
            A *dict* containing entries for all state variables.

        """
        for attr, value in state.items():
            setattr(self.state, attr, value)


def create_heatdemand(config, state):
    """Creates a household object"""

    if not DEMAND_PROFILES:
        load_demand_profiles()

    p_th_prod = -config.get("chp_p_th_prod_kw", -400)
    profile = find_demand_profile(p_th_prod)

    # adapt consumer constant of the household, so that power and
    # consumption better fit to each other
    profile["consumer_constant"] = p_th_prod * 24 * 2 / 3

    heatdemand_obj = HeatDemand(profile, state)

    return heatdemand_obj


def load_demand_profiles():
    """Write me"""
    global DEMAND_PROFILES
    DEMAND_PROFILES = dict()
    for name, func in hprofiles.__dict__.items():
        if "__" not in name:
            DEMAND_PROFILES[name] = func()


def find_demand_profile(chp_p_th_kw):
    """Write me"""
    min_consumption = dict()
    max_consumption = dict()

    for key, val in DEMAND_PROFILES.items():
        # average annual consumption per average operating
        # hours of the chp
        consumption = (
            val["consumer_constant"] * 365 / HeatDemand.CHP_AVERAGE_RUNTIME
        )
        # calculate lower power boundary
        min_consumption[key] = consumption * HeatDemand.CHP_MIN_TH_CONS_COEF
        # calculate upper power boundary
        max_consumption[key] = consumption * HeatDemand.CHP_MAX_TH_CONS_COEF

    possible_profiles = list()
    distances = dict()

    # Look for fitting households
    for key, val in DEMAND_PROFILES.items():
        if abs(chp_p_th_kw >= min_consumption[key]) and abs(
            chp_p_th_kw <= max_consumption[key]
        ):
            # Profile fits optimally
            possible_profiles.append(key)
            distances[key] = 0.0
        else:
            # Profile does not fit, calculate difference
            lower = abs(abs(chp_p_th_kw) - min_consumption[key])
            upper = abs(abs(chp_p_th_kw) - max_consumption[key])
            distances[key] = lower if lower < upper else upper

    if len(possible_profiles) == 1:
        # Exactly one matching profile
        demand_profile = DEMAND_PROFILES[possible_profiles[0]].copy()
    elif len(possible_profiles) > 1:
        # More than one matching profile, select randomly
        demand_profile = DEMAND_PROFILES[
            random.choice(possible_profiles)
        ].copy()
    else:
        # No matching profile, select the nearest one

        min_type = "one_family"
        min_val = float("Inf")

        for key, val in distances.items():
            if val < min_val:
                min_val = val
                min_type = key

        demand_profile = DEMAND_PROFILES[min_type].copy()
    return demand_profile



"""This module contains configurations for chp models.

Example:

    from pysimmods.chpsystemsim import CHPSystem
    chp = CHPSystem(chp_params(7), chp_init(7))

"""
import sys


def chp_preset(p_kw, start_date="2017-01-01 00:00:00+0100", **kwargs):
    """Return the parameter configuration for a chp model
    from the pysimmods package.

    """
    thismodule = sys.modules[__name__]

    if p_kw in (7, 14, 200, 400):
        pmethod = f"_chp_{int(p_kw)}kw_params"
        imethod = f"_chp_{int(p_kw)}kw_init"
    else:
        raise ValueError("Not supported model size {} p_kw".format(p_kw))

    params = getattr(thismodule, pmethod)()
    params["sign_convention"] = "active"
    params["flip_e_th_demand_sign"] = True
    inits = {
        "chp": getattr(thismodule, imethod)(),
        "household": {
            "start_date": start_date,
            "t_last_3": 2.1,
            "t_last_2": 1.7,
            "t_last_1": 4.2,
        },
    }
    return params, inits


def _chp_7kw_params():
    params = {
        "chp": {
            "p_max_kw": 7.0,
            "p_min_kw": 3.5,
            "p_2_p_th_percent": 257.0,
            "eta_max_percent": 86.1,
            "eta_min_percent": 80.5,
            "own_consumption_kw": 0.55,
            "active_min_s": 0.0,
            "inactive_min_s": 0.0,
            "lubricant_max_l": 10.0,
            "lubricant_ml_per_h": 10.0,
            "storage_cap_l": 1050.0,
            "storage_consumption_kwh_per_day": 0.75,
            "storage_t_min_c": 40.0,
            "storage_t_max_c": 85.0,
        }
    }
    return params


def _chp_7kw_init():
    init_vals = {
        "lubricant_l": 9,
        "active_s": 1800,
        "inactive_s": 0,
        "is_active": True,
        "storage_t_c": 50,
    }
    return init_vals


def _chp_14kw_params():
    params = {
        "chp": {
            "p_max_kw": 14.0,
            "p_min_kw": 7,
            "p_2_p_th_percent": 248.0,
            "eta_max_percent": 87.2,
            "eta_min_percent": 78.4,
            "own_consumption_kw": 0.588,
            "active_min_s": 0.0,
            "inactive_min_s": 0.0,
            "lubricant_max_l": 14.0,
            "lubricant_ml_per_h": 12.5,
            "storage_cap_l": 2100.0,
            "storage_consumption_kwh_per_day": 0.77,
            "storage_t_min_c": 40.0,
            "storage_t_max_c": 85.0,
        }
    }
    return params


def _chp_14kw_init():
    init_vals = {
        "lubricant_l": 9,
        "active_s": 1800,
        "inactive_s": 0,
        "is_active": True,
        "storage_t_c": 50,
    }
    return init_vals


def _chp_200kw_params():
    params = {
        "chp": {
            "p_max_kw": 198,
            "p_min_kw": 99,
            "p_2_p_th_percent": 148.0,
            "eta_max_percent": 88.8,
            "eta_min_percent": 87.1,
            "own_consumption_kw": 2.9,
            "active_min_s": 0.0,
            "inactive_min_s": 0.0,
            "lubricant_max_l": 95.0,
            "lubricant_ml_per_h": 80.0,
            "storage_cap_l": 30_000.0,
            "storage_consumption_kwh_per_day": 1.63,
            "storage_t_min_c": 40.0,
            "storage_t_max_c": 85.0,
        }
    }
    return params


def _chp_200kw_init():
    init_vals = {
        "lubricant_l": 90,
        "active_s": 1800,
        "inactive_s": 0,
        "is_active": True,
        "storage_t_c": 50,
    }
    return init_vals


def _chp_400kw_params():
    params = {
        "chp": {
            "p_max_kw": 400,
            "p_min_kw": 200,
            "p_2_p_th_percent": 128.0,
            "eta_max_percent": 88.7,
            "eta_min_percent": 87.4,
            "own_consumption_kw": 5.3,
            "active_min_s": 0.0,
            "inactive_min_s": 0.0,
            "lubricant_max_l": 155.0,
            "lubricant_ml_per_h": 130.0,
            "storage_cap_l": 60_000.0,
            "storage_consumption_kwh_per_day": 2.13,
            "storage_t_min_c": 40.0,
            "storage_t_max_c": 85.0,
        }
    }
    return params


def _chp_400kw_init():
    init_vals = {
        "lubricant_l": 154,
        "active_s": 1800,
        "inactive_s": 0,
        "is_active": True,
        "storage_t_c": 50,
    }
    return init_vals



def diesel_presets(p_max_kw):
    params = _get_dict(p_max_kw)
    inits = {"p_kw": 0, "q_kvar": 0}

    params["sign_convention"] = "active"

    return params, inits


def _get_dict(p_max_kw):
    params = {
        "p_max_kw": p_max_kw,
        "p_min_kw": 0,
        "q_max_kvar": 0,
        "q_min_kvar": 0,
    }
    return params



"""This module contains the config model for the chp."""
from pysimmods.model.config import ModelConfig


class CHPCNGConfig(ModelConfig):
    """A CHP CNG configuration parameter class."""

    def __init__(self, params):
        super().__init__(params)

        self.pn_stages_kw = params["pn_stages_kw"]
        self.pn_stages_kw = [abs(stage) for stage in self.pn_stages_kw]
        """Possible electrial power stages in [kW].
        The power level 0 (unit off) should not be included.
        """
        self.p_max_kw = max(self.pn_stages_kw)
        self.p_min_kw = min(self.pn_stages_kw)

        self.eta_stages_percent = params["eta_stages_percent"]
        """Electrical efficiency at each power stage in [%]."""
        self.eta_th_stages_percent = params["eta_th_stages_percent"]
        """Thermal efficiency at each power state in [%]."""

        self.restarts_per_day = params["restarts_per_day"]
        """Allowed restarts per day."""
        self.active_min_s = params["active_min_s"]
        """Minimal active time of the unit in [s]."""
        self.active_max_s_per_day = params["active_max_s_per_day"]
        """Maximal cumulated active time per day in [s]."""
        self.inactive_min_s = params["inactive_min_s"]
        """Minimal time to be inactive in [s]."""
        self.inactive_max_s_per_day = params["inactive_max_s_per_day"]
        """Maximal cumulated inactive time per day in [s]."""

        self.e_ch4_kwh = 9.94
        """Energy of methane cas in [kWh]."""
        self.ch4_concentration_percent = params.get(
            "ch4_concentration_percent", 50.302
        )
        """Concentration of methane gas in [%]."""

        self.default_schedule = [50] * 24



"""This module contains a model of a pv system with pv modules and an
inverter"""
from pysimmods.generator.pvsim.pvp import PhotovoltaicPowerPlant
from pysimmods.generator.pvsystemsim.config import PVSystemConfig
from pysimmods.generator.pvsystemsim.inputs import PVSystemInputs
from pysimmods.generator.pvsystemsim.state import PVSystemState
from pysimmods.model.qgenerator import QGenerator
from pysimmods.other.invertersim.inverter import Inverter


class PVPlantSystem(QGenerator):
    """Pv system with pv modules and inverter"""

    def __init__(self, params, inits):
        self.config = PVSystemConfig(params)
        self.inputs = PVSystemInputs()
        self.state = PVSystemState(inits)
        self.pv = PhotovoltaicPowerPlant(params["pv"], inits["pv"])
        self.inverter = Inverter(params["inverter"])

    def step(self):
        """Perform simulation step"""

        # Step the pv plant
        self.pv.inputs.bh_w_per_m2 = self.inputs.bh_w_per_m2
        self.pv.inputs.dh_w_per_m2 = self.inputs.dh_w_per_m2
        self.pv.inputs.s_module_w_per_m2 = self.inputs.s_module_w_per_m2
        self.pv.inputs.t_air_deg_celsius = self.inputs.t_air_deg_celsius
        self.pv.inputs.step_size = self.inputs.step_size
        self.pv.inputs.now_dt = self.inputs.now_dt

        self.pv.step()

        # Step the inverter
        self.inverter.inputs.p_in_kw = self.pv.state.p_kw
        self.inverter.inputs.p_set_kw = self.inputs.p_set_kw
        self.inverter.inputs.q_set_kvar = self.inputs.q_set_kvar
        self.inverter.inputs.cos_phi_set = self.inputs.cos_phi_set
        self.inverter.inputs.inductive = self.inputs.inverter_inductive

        self.inverter.step()

        # Update state
        self.state.t_module_deg_celsius = self.pv.state.t_module_deg_celsius
        self.state.p_kw = self.inverter.state.p_kw
        self.state.q_kvar = self.inverter.state.q_kvar
        self.state.cos_phi = self.inverter.state.cos_phi
        self.state.inverter_inductive = self.inverter.state.inductive

        self.inputs.reset()

    def get_state(self):
        """Get state"""
        state_dict = {
            "pv": self.pv.get_state(),
            "inverter": self.inverter.get_state(),
        }
        return state_dict

    def set_state(self, state_dict):
        """Set state"""
        self.pv.set_state(state_dict["pv"])
        self.inverter.set_state(state_dict["inverter"])

        self.state.t_module_deg_celsius = self.pv.state.t_module_deg_celsius
        self.state.p_kw = self.inverter.state.p_kw
        self.state.q_kvar = self.inverter.state.q_kvar

    def set_q_kvar(self, q_kvar: float) -> None:
        self.inputs.q_set_kvar = q_kvar



#!/usr/bin/env python3
"""Setup file for pysimmods package."""
import setuptools

with open("VERSION") as freader:
    VERSION = freader.readline().strip()


with open("README.md") as freader:
    README = freader.read()

install_requirements = [
    "numpy",
    "pandas",
    "mosaik_api",
]
development_requirements = {
    "tox",
    "pytest",
    "flake8",
    "rope",
    "twine",
    "black==22.3.0",
    "coverage",
    "sphinx",
    "pytest-cov",
    "matplotlib",
    "wheel",
    "midas-util>=1.0.0",
}

extras = {"dev": development_requirements}

setuptools.setup(
    name="pysimmods",
    version=VERSION,
    author="Stephan Balduin",
    author_email="stephan.balduin@offis.de",
    description="A set of simulation models representing different "
    "distributed consumer or generator units.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/midas-mosaik/pysimmods",
    packages=setuptools.find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=install_requirements,
    extras_require=extras,
    license="LGPL",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: "
        "GNU Lesser General Public License v2 (LGPLv2)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.8",
)
